/*
 * Multi precision toolbox for Scilab
 * Copyright (C) 2009 - Jonathan Blanchard
 *
 * This file must be used under the terms of the CeCILL.
 * This source file is licensed as described in the file COPYING, which
 * you should have received as part of this distribution.  The terms
 * are also available at
 * http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
 */

#include "MALLOC.h"
#include "sciprint.h"
#include "stack-c.h"
#include "api_common.h"
#include "api_double.h"
#include "api_string.h"

#include "scimps.h"

/*
 * Element-wise division.
 * Scilab definition :
 * mps_add( rop, op1, op2, [rnd] )
 * rop : Result operand, multi-precision matrix.
 * op1, op2 : Matrices to be added, Double or multi-precision matrices.
 * rnd : Optional rounding mode. One character from the following :
 * 'n' : Round to nearest
 * 'z' : Round toward 0 (Truncate)
 * 'u' : Round toward +infinity
 * 'd' : Round toward -infinity
 *
 * Notes :
 * Currently op1 and op2 cannot both be matrices of double.
 * rop must be of the correct dimensions as it will be used as-is for the result.
 */
int sci_mps_div( char *fname, unsigned long fname_len )
{
    int *arg1, *arg2, *arg3, *arg4;
    int m1, m2, n1, n2;
    int typearg2, typearg3;
    mps_ptr rop, op1, op2;
    double *dptr1, *dptr2;
    mp_rnd_t rnd = GMP_RNDN;
    int rnum;
    char rndchar[2];
    int pilen[1];

    arg4 = NULL;

    /* Check the number of lhs and rhs arguments. */
    CheckRhs(3,4);
    CheckLhs(1,1);

    /* Get address of inputs */
    getVarAddressFromPosition( pvApiCtx, 1, &arg1 );
    getVarAddressFromPosition( pvApiCtx, 2, &arg2 );
    getVarAddressFromPosition( pvApiCtx, 3, &arg3 );

    /* If there is a fourth argument, get it. */
    if( Rhs == 4 )
        getVarAddressFromPosition( pvApiCtx, 4, &arg4 );

    /* Check the size and type of the arguments. */
    rnum = MpsCheck1( fname, arg1, arg2, arg3, arg4 );

    if( rnum != 0 )
        return 0;

    /* Get the rounding mode. */
    if( Rhs == 4 )
    {
        GetRhsStringVar( 4 , &m1, &n1, pilen, rndchar );
        if( rndchar[0] == 'n' )
            rnd = GMP_RNDN;
        else if( rndchar[0] == 'z' )
            rnd = GMP_RNDZ;
        else if( rndchar[0] == 'u')
            rnd = GMP_RNDU;
        else if( rndchar[0] == 'd' )
            rnd = GMP_RNDD;
        else
        {
            sciprint("%s: Invalid rounding mode specified. Valid options : 'n', 'z', 'u', 'd'.\n",fname);
            return 0;
        }
    }

    /* Fetch the result operand. */
    rop = MpsGetMatrix( 1 );

    getVarType( pvApiCtx, arg2, &typearg2 );
    getVarType( pvApiCtx, arg3, &typearg3 );

	/* Find which situation we are in and jump to the corresponding case. */
    switch (typearg2)
    {
        case sci_matrix:
            switch (typearg3)
            {
                case sci_matrix:
                    goto DOUBLE_DOUBLE;
                case sci_mlist:
                    goto DOUBLE_TLIST;
                default:
                    return 0;
    		}
        case sci_mlist:
            switch (typearg3)
            {
                case sci_matrix:
                    goto TLIST_DOUBLE;
                case sci_mlist:
                    goto TLIST_TLIST;
                default:
                    return 0;
            }
        default:
            return 0;
    }

    /* Never place something here...*/

    DOUBLE_DOUBLE: /* (double*, double*) */

        getMatrixOfDouble( pvApiCtx, arg2, &m1, &n1, &dptr1 );
        getMatrixOfDouble( pvApiCtx, arg3, &m2, &n2, &dptr2 );

        if( (m1 == 1) && (n1 == 1) )
            mps_scalar_double_div_double( rop, dptr1[0], dptr2, 0, rnd );
        else if( (m2 == 1) && (n2 == 1) )
            mps_double_div_scalar_double( rop, dptr1, 0, dptr2[0], rnd );
        else
            mps_double_div_double( rop, dptr1, 0, dptr2, 0, rnd );

    goto END;

    DOUBLE_TLIST: /* (double*, mps_t) */

        getMatrixOfDouble( pvApiCtx, arg2, &m1, &n1, &dptr1 );
        op2 = MpsGetMatrix( 3 );

        /*
         * Check if we are performing a scalar-matrix operation and call the
         * required mps function.
         */
        if( (m1 == 1) && (n1 == 1) )
            mps_scalar_double_div( rop, dptr1[0], op2, rnd);
        else if( (MPS_NUMCOL(op2) == 1) && (MPS_NUMROW(op2) == 1) )
        {
            mps_double_div_scalar( rop, dptr1, 0, mps_get_ele(op2, 1, 1), rnd );
            return 0;
        }
        else
            mps_double_div( rop, dptr1, op2, 0, rnd );

    goto END;

    TLIST_DOUBLE: /* (mps_t, double*) */

        op1 = MpsGetMatrix( 2 );
        getMatrixOfDouble( pvApiCtx, arg3, &m2, &n2, &dptr2 );

        if( (m2 == 1) && (n2 == 1) )
            mps_div_scalar_double( rop, op1, dptr2[0], rnd );
        else if( (MPS_NUMCOL(op1) == 1) && (MPS_NUMROW(op1) == 1) )
        {
            mps_scalar_div_double( rop, mps_get_ele(op1, 1, 1), dptr2, 0, rnd );
            return 0;
        }
        else
            mps_div_double( rop, op1, dptr2, 0, rnd );

    goto END;

    TLIST_TLIST: /* (mps_t, mps_t) */

        op1 = MpsGetMatrix( 2 );
        op2 = MpsGetMatrix( 3 );

        if( (MPS_NUMCOL(op1) == 1) && (MPS_NUMROW(op1) == 1) )
            mps_scalar_div( rop, mps_get_ele(op1, 1, 1), op2, rnd );
        else if( (MPS_NUMCOL(op2) == 1) && (MPS_NUMROW(op2) == 1) )
            mps_div_scalar( rop, op1, mps_get_ele(op2, 1, 1), rnd );
        else
            mps_div( rop, op1, op2, rnd );

    END:

    return 0;
}

