/*
 * Multi precision toolbox for Scilab
 * Copyright (C) 2009 - Jonathan Blanchard
 *
 * This file must be used under the terms of the CeCILL.
 * This source file is licensed as described in the file COPYING, which
 * you should have received as part of this distribution.  The terms
 * are also available at
 * http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
 */

#include "MALLOC.h"
#include "sciprint.h"
#include "stack-c.h"
#include "api_common.h"
#include "api_double.h"
#include "api_string.h"

#include "scimps.h"

/*
 * Element-wise power.
 * Scilab definition :
 * mps_pow( rop, op1, op2, [rnd] )
 * rop : Result operand, multi-precision matrix.
 * rnd : Optional rounding mode. One character from the following :
 * 'n' : Round to nearest
 * 'z' : Round toward 0 (Truncate)
 * 'u' : Round toward +infinity
 * 'd' : Round toward -infinity
 */
int sci_mps_pow( char *fname, unsigned long fname_len )
{
    int *arg1, *arg2, *arg3, *arg4;
    int m1, m2, n1, n2;
    int typearg1, typearg2, typearg3;
    mps_ptr rop, op1, op2;
    mpfr_t mpfrtmp;
    double *dptr1, *dptr2;
    mp_rnd_t rnd = GMP_RNDN;
    int rnum;

    arg4 = NULL;

    /* Check the number of lhs and rhs arguments. */
    CheckRhs(3,4);
    CheckLhs(1,1);

    /* Get address of inputs */
    getVarAddressFromPosition( pvApiCtx, 1, &arg1 );
    getVarAddressFromPosition( pvApiCtx, 2, &arg2 );
    getVarAddressFromPosition( pvApiCtx, 3, &arg3 );

    getVarType( pvApiCtx, arg1, &typearg1 );

    if( typearg1 != sci_mlist )
    {
        sciprint( "%s: Wrong type for argument 1. Multi-precision matrix expected.\n", fname );
        return 0;        
    }

    /* If there is a 4th argument fetch it. */
    if( Rhs == 4 )
    {
        getVarAddressFromPosition( pvApiCtx, 4, &arg4 );
        rnum = MpsGetRoundingArg( fname, &rnd, arg4, 4 );
        if( rnum != 0 )
            return 0;
    }

    /* Fetch the result operand. */
    rop = MpsGetMatrix( 1 );

    getVarType( pvApiCtx, arg2, &typearg2 );
    getVarType( pvApiCtx, arg3, &typearg3 );

	/* Find which situation we are in and jump to the corresponding case. */
    switch (typearg2)
    {
        case sci_matrix:
            switch (typearg3)
            {
                case sci_matrix:
                    goto DOUBLE_DOUBLE;
                case sci_mlist:
                    goto DOUBLE_TLIST;
                default:
                    return 0;
    		}
        case sci_mlist:
            switch (typearg3)
            {
                case sci_matrix:
                    goto TLIST_DOUBLE;
                case sci_mlist:
                    goto TLIST_TLIST;
                default:
                    return 0;
            }
        default:
            return 0;
    }

    /* Never place something here...*/

    DOUBLE_DOUBLE: /* (double*, double*) */

        getMatrixOfDouble( pvApiCtx, arg2, &m1, &n1, &dptr1 );
        getMatrixOfDouble( pvApiCtx, arg3, &m2, &n2, &dptr2 );

        sciprint( "%s: Wrong type for argument 2. Multi-precision matrix expected.\n", fname );
        return 0;

    goto END;

    DOUBLE_TLIST: /* (double*, mps_t) */

        sciprint( "%s: Wrong type for argument 2. Multi-precision matrix expected.\n", fname );
        return 0;

    goto END;

    TLIST_DOUBLE: /* (mps_t, double*) */

        op1 = MpsGetMatrix( 2 );
        getMatrixOfDouble( pvApiCtx, arg3, &m2, &n2, &dptr2 );

        if( (m2 != 1) || (n2 != 1) )
        {
            sciprint( "%s: Wrong size for argument 3. Scalar expected.\n", fname );
            return 0;
        }

        if( MPS_NUMCOL(rop) != MPS_NUMCOL(op1) || MPS_NUMROW(rop) != MPS_NUMROW(op1) )
        {
            sciprint( "%s: Inconsistent matrix operation, input and out operands differ in size.\n", fname );
            return 0;
        }
        
        mpfr_init2( mpfrtmp, MPS_PREC(rop) );
        mpfr_set_d( mpfrtmp, dptr2[0], GMP_RNDN );

        mps_pow( rop, op1, mpfrtmp, rnd );

        mpfr_clear( mpfrtmp );

    goto END;

    TLIST_TLIST: /* (mps_t, mps_t) */

        op1 = MpsGetMatrix( 2 );
        op2 = MpsGetMatrix( 3 );

        if( (MPS_NUMCOL(op2) != 1) || (MPS_NUMROW(op2) != 1) )
        {
            sciprint( "%s: Wrong size for argument 3. Scalar expected.\n", fname );
            return 0;
        }

        if( MPS_NUMCOL(rop) != MPS_NUMCOL(op1) || MPS_NUMROW(rop) != MPS_NUMROW(op1) )
        {
            sciprint( "%s: Inconsistent matrix operation, input and out operands differ in size.\n", fname );
            return 0;
        }

        mps_pow( rop, op1, mps_get_ele(op2,1,1), rnd );

    END:

    return 0;
}
